/**
 * An implementation of a AVL Tree Data structure for integers.
 */
class AVLTree2 {

	// constant used for toString formatting
	private static final String INDENT = "   ";

	/**
	 * The root node of the represented AVL tree.
	 */
	private Node root;

	/**
	 * Creates an empty AVL tree.
	 */
	public AVLTree2() {
		// in the beginning, the tree is empty
		root = null;
	}

	/**
	 * Checks whether this tree is empty or not.
	 * 
	 * @return true if empty, false otherwise.
	 */
	public boolean isEmpty() {
		return root == null;
	}

	/**
	 * Tests whether this AVL tree contains the given element.
	 * 
	 * @param element
	 *            the element to be searched
	 * @return true if contained, false otherwise
	 */
	public boolean contains(int element) {
		// TODO: implement this operation by navigating through 
		// the nodes starting at the root node
		return false;
	}

	/**
	 * Inserts the given element inside this tree.
	 * 
	 * @param value
	 *            the value to be inserted
	 */
	public void insert(int value) {
		// TODO: implement this operation.
		// see slides 157 - 159 of the handout
		// you might wanna use balGrow(..) to handle the balancing
	}

	/**
	 * React on the grew of the given sub-tree.
	 * 
	 * @param n
	 *            the root node of the sub-tree which grew by one level.
	 */
	private void balGrow(Node n) {
		assert (n.hdiff == -1 || n.hdiff == 1);

		// get parent node
		Node p = n.parent;
		
		// TODO: implement this the rest of this method
		// reacting on a grown subtree n by:
		//  - studying slides 160 - 165
		//  - apply case distinction to decide whether rotations are required
		//	- updating hdiff of the parent node p
		//	- apply rotations if necessary (Single or Double rotations)
		//	- update hdiff values of nodes involved in rotations
		// 
		// Hint: this might require some cumbersome case distinction code
	}

	/**
	 * This method is applying a left-rotation on the given node. His right
	 * child will become the new root node of the corresponding sub-tree.
	 * 
	 * @param n
	 *            the node to be rotated toward the left.
	 */
	private void rotateLeft(Node p) {
		Node n = p.right;

		// TODO: implement the left-rotation of node p
		//  - study slide 163 (this rotation is just the inverse direction)
		//	- perform the necessary local reference modifications
		//	- do not forget the parent pointers!
		//	- the hdiff values should be updated by the caller => to increase reusability
	}

	/**
	 * This method is applying a right-rotation on the given node. His left
	 * child will become the new root node of the corresponding sub-tree.
	 * 
	 * @param n
	 *            the node to be rotated toward the right.
	 */
	private void rotateRight(Node p) {
		Node n = p.left;

		// TODO: implement the left-rotation of node p
		//  - study slide 163
		//	- perform the necessary local reference modifications
		//	- do not forget the parent pointers!
		//	- the hdiff values should be updated by the caller => to increase reusability
	}

	/**
	 * Allows to print this AVL tree to some output stream (e.g. console).
	 */
	@Override
	public String toString() {

		// check for an empty tree
		if (isEmpty()) {
			return "Empty AVL Tree";
		}

		// build up a more complex string representation containing some details
		StringBuilder builder = new StringBuilder();
		appendSubTree(builder, root, "");

		// let the builder compose the string and done
		return builder.toString();
	}

	/**
	 * A utility function allowing to recursively print an AVL sub-tree to some
	 * string builder.
	 * 
	 * @param builder
	 *            the string builder to be used for generating the output
	 * @param root
	 *            the root of the sub-tree to be printed
	 * @param prefix
	 *            the prefix to be used to realize indentation
	 */
	private void appendSubTree(StringBuilder builder, Node root, String prefix) {

		// terminal case: root is null
		if (root == null) {
			return; // nothing to do
		}

		String newPrefix = prefix + INDENT;

		// append left side (recursively, with increased prefix)
		appendSubTree(builder, root.left, newPrefix);

		// append current node
		builder.append(prefix).append(root.value);

		// add the hdiff value and the actual difference for debugging reasons
		builder.append("(");
		builder.append(root.hdiff);
		builder.append("/");
		builder.append(getHeight(root.right) - getHeight(root.left));
		builder.append(")");

		// finish this line
		builder.append("\n");

		// append right side
		appendSubTree(builder, root.right, newPrefix);
	}

	/**
	 * Checks whether the represented AVL Tree is valid.
	 * 
	 * @return true if so, false otherwise.
	 */
	public boolean checkAVLTree() {
		// check special properties of the root nodes and the rest of the nodes
		// recursively
		return root == null || (checkNode(root) && root.parent == null);
	}

	/**
	 * Checks whether the sub-tree rooted by the given node is a valid AVL tree.
	 * 
	 * @param cur
	 *            the root of the sub-tree to be tested
	 * @return true if valid, false otherwise
	 */
	private boolean checkNode(Node cur) {
		// empty sub-tree is valid
		if (cur == null) {
			return true;
		}

		// check invariants on the current sub-tree
		boolean res = true;

		// height difference must be right
		res = res && cur.hdiff == getHeight(cur.right) - getHeight(cur.left);

		// left side values must be smaller and parent links must be correct
		if (cur.left != null) {
			res = res && cur.left.value < cur.value;
			res = res && cur.left.parent == cur;
		}
		if (cur.right != null) {
			res = res && cur.value < cur.right.value;
			res = res && cur.right.parent == cur;
		}

		res = res && checkNode(cur.left) && checkNode(cur.right);
		return res;
	}

	/**
	 * Obtains the height of a sub-tree rooted by the given node.
	 * 
	 * @param subtree
	 *            the root of the sub-tree which's height should be obtained
	 * @return the height of the requested sub-tree
	 */
	private int getHeight(Node subtree) {
		// terminal case
		if (subtree == null) {
			return 0;
		}
		// step case (divide & conquer)
		return Math.max(getHeight(subtree.left), getHeight(subtree.right)) + 1;
	}

	/**
	 * A helper class used to represent nodes within this AVL Tree.
	 */
	private class Node {

		/**
		 * The value to be stored by this node.
		 */
		final int value;

		/**
		 * The height difference of the left and right sub-tree.
		 */
		int hdiff = 0;

		/**
		 * The parent of this node.
		 */
		Node parent;

		/**
		 * The left sub-tree.
		 */
		Node left = null;

		/**
		 * The right sub-tree.
		 */
		Node right = null;

		/**
		 * A constructor for a node storing the given value.
		 * 
		 * @param value
		 *            the value to be stored by the new node
		 * @param parent
		 *            the parent node of the new node.
		 */
		public Node(int value, Node parent) {
			this.value = value;
			this.parent = parent;
		}

		@Override
		public String toString() {
			return "" + value;
		}
	}
}
