import java.util.ArrayList;
import java.util.Collections;

/**
 * This class implements a main program allowing to test the AVL Tree implementation. 
 */
class TestCase {
	
	public static void main(String[] args) {
	
		// create a Tree handling Integers (~int)
		AVLTree<Integer> tree = new AVLTree<Integer>();
		
		System.out.println("Begin:");
		System.out.println(tree);
		
		for(int i : new int[]{1,7,5,9,6,8,3,4,2}) {
			
			tree.insert(i);
			System.out.println();
			System.out.println("After inserting " + i + ": ");
			System.out.println("Tree contains " + i + ": " + tree.contains(i));
			System.out.println(tree);
			checkTree(tree);
		}
		
		// the same for strings ...
		AVLTree<String> tree2 = new AVLTree<String>();
		for(String i : new String[]{"A", "L", "G", "O", "R", "I", "T", "H", "M"}) {
			
			tree2.insert(i);
			System.out.println();
			System.out.println("After inserting " + i + ": ");
			System.out.println("Tree contains " + i + ": " + tree2.contains(i));
			System.out.println(tree2);
			checkTree(tree2);
		}
		
		// ... and a final stress test
		
		// generate a random permutation of 0 .. N-1
		final int N = 4000;
		ArrayList<Integer> data = new ArrayList<Integer>(N);
		for (int j=0; j<N; j++) {
			data.add(j);
		}
		
		for (int i=0; i<10; i++) {	
			Collections.shuffle(data);
			
			System.out.println("Running random run " + (i+1) + " ...");
			
			// insert it into the tree
			AVLTree<Integer> tree3 = new AVLTree<Integer>();
			for(int cur : data) {
				tree3.insert(cur);
				checkTree(tree3);
			}
			
			// check whether all elements are present
			for (int cur : data) {
				if (!tree3.contains(cur)) {
					System.err.println("Tree does not contain " + cur);
					return;
				}
			}
		}
		
		// nothing failed
		System.out.println("Congratulation - all test passed!");
	}
	
	private static void checkTree(AVLTree<?> tree) {
		if (!tree.checkAVLTree()) {
			System.err.println("ERROR: invalid AVL tree");
			System.err.println(tree);
			System.exit(1);
		}
	}
}